/*****************************************************************************************
 *                                                                                       *
 * OpenSpace                                                                             *
 *                                                                                       *
 * Copyright (c) 2014-2022                                                               *
 *                                                                                       *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this  *
 * software and associated documentation files (the "Software"), to deal in the Software *
 * without restriction, including without limitation the rights to use, copy, modify,    *
 * merge, publish, distribute, sublicense, and/or sell copies of the Software, and to    *
 * permit persons to whom the Software is furnished to do so, subject to the following   *
 * conditions:                                                                           *
 *                                                                                       *
 * The above copyright notice and this permission notice shall be included in all copies *
 * or substantial portions of the Software.                                              *
 *                                                                                       *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,   *
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A         *
 * PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT    *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF  *
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE  *
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                                         *
 ****************************************************************************************/

/*****************************************************************************************
 * Modified parts of the code (4D texture mechanism, analytical transmittance etc)       *
 * from Eric Bruneton is used in the following code.                                     * 
 ****************************************************************************************/

/**
 * Precomputed Atmospheric Scattering
 * Copyright (c) 2008 INRIA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, this list of
 *    conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list
 *    of conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


const int INSCATTER_INTEGRAL_SAMPLES = 50;
const float M_PI = 3.141592657;
const float ATM_EPSILON = 1.0;

// In the following shaders r (altitude) is the length of vector/position x in the
// atmosphere (or on the top of it when considering an observer in space), where the light
// is coming from the opposite direction of the view direction, here the vector v or
// viewDirection. Rg is the planet radius and Rt the atmosphere radius.

// Calculate the distance of the ray starting at x (height r) until the planet's ground
// or top of atmosphere
// r := || vec(x) || e [0, Rt]
// mu := cosine of the zeith angle of vec(v). Or mu = (vec(x) * vec(v))/r 
float rayDistance(float r, float mu, float Rt, float Rg) {
  // The light ray starting at the observer in/on the atmosphere can have to possible end
  // points: the top of the atmosphere or the planet ground. So the shortest path is the
  // one we are looking for, otherwise we may be passing through the ground
  
  // cosine law
  float atmRadiusEps2 = (Rt + ATM_EPSILON) * (Rt + ATM_EPSILON);
  float mu2 = mu * mu;
  float r2 = r * r;
  float rayDistanceAtmosphere = -r * mu + sqrt(r2 * (mu2 - 1.0) + atmRadiusEps2); 
  float delta = r2 * (mu2 - 1.0) + Rg*Rg;

  // Ray may be hitting ground
  if (delta >= 0.0) {
    float rayDistanceGround = -r * mu - sqrt(delta);
    if (rayDistanceGround >= 0.0) {
      return min(rayDistanceAtmosphere, rayDistanceGround);
    }
  }
  return rayDistanceAtmosphere;
}

// Given the windows's fragment coordinates, for a defined view port, gives back the
// interpolated r e [Rg, Rt] and mu, muSun amd nu e [-1, 1]
// r := height of starting point vect(x)
// mu := cosine of the zeith angle of vec(v). Or mu = (vec(x) * vec(v))/r
// muSun := cosine of the zeith angle of vec(s). Or muSun = (vec(s) * vec(v))
// nu := cosone of the angle between vec(s) and vec(v)
// dhdH := it is a vec4. dhdH.x stores the dminT := Rt - r, dhdH.y stores the dH value
//         (see paper), dhdH.z stores dminG := r - Rg and dhdH.w stores dh (see paper)
void unmappingMuMuSunNu(float r, vec4 dhdH, int SAMPLES_MU, float Rg, float Rt,
                        int SAMPLES_MU_S, int SAMPLES_NU,
                        out float mu, out float muSun, out float nu)
{
  // Window coordinates of pixel (uncentering also)
  vec2 fragment = gl_FragCoord.xy - vec2(0.5);

  // Pre-calculations
  float r2  = r * r;
  float Rg2 = Rg * Rg;
  
  float halfSAMPLE_MU = float(SAMPLES_MU) / 2.0;
  // If the (vec(x) dot vec(v))/r is negative, i.e., the light ray has great probability
  // to touch the ground, we obtain mu considering the geometry of the ground
  if (fragment.y < halfSAMPLE_MU) {
    float ud = 1.0 - (fragment.y / (halfSAMPLE_MU - 1.0));
    float d = min(max(dhdH.z, ud * dhdH.w), dhdH.w * 0.999);
    // cosine law: Rg^2 = r^2 + d^2 - 2rdcos(pi-theta) where cosine(theta) = mu
    mu = (Rg2 - r2 - d * d) / (2.0 * r * d);
    // We can't handle a ray inside the planet, i.e., when r ~ Rg, so we check against it.
    // If that is the case, we approximate to a ray touching the ground.
    // cosine(pi-theta) = dh/r = sqrt(r^2-Rg^2)
    // cosine(theta) = - sqrt(1 - Rg^2/r^2)
    mu = min(mu, -sqrt(1.0 - (Rg2 / r2)) - 0.001);
  }
  // The light ray is touching the atmosphere and not the ground
  else {
    float d = (fragment.y - halfSAMPLE_MU) / (halfSAMPLE_MU - 1.0);
    d = min(max(dhdH.x, d * dhdH.y), dhdH.y * 0.999);
    // cosine law: Rt^2 = r^2 + d^2 - 2rdcos(pi-theta) where cosine(theta) = mu
    mu = (Rt*Rt - r2 - d * d) / (2.0 * r * d);
  }
  
  float modValueMuSun = mod(fragment.x, float(SAMPLES_MU_S)) / (float(SAMPLES_MU_S) - 1.0);
  // The following mapping is different from the paper. See Collienne for an details.
  muSun = tan((2.0 * modValueMuSun - 1.0 + 0.26) * 1.1) / tan(1.26 * 1.1);
  nu = -1.0 + floor(fragment.x / float(SAMPLES_MU_S)) / (float(SAMPLES_NU) - 1.0) * 2.0;
}

// Function to access the transmittance texture. Given r and mu, returns the transmittance
// of a ray starting at vec(x), height r, and direction vec(v), mu, and length until it
// hits the ground or the top of atmosphere.
// r := height of starting point vect(x)
// mu := cosine of the zeith angle of vec(v). Or mu = (vec(x) * vec(v))/r
vec3 transmittance(sampler2D tex, float r, float mu, float Rg, float Rt) {
  float H2 = Rt*Rt - Rg*Rg;
  float u_r = sqrt((r*r - Rg*Rg) / H2);

  // Old version
  // -----------
  // Given the position x (here the altitude r) and the view angle v
  // (here the cosine(v)= mu), we map this
  // float u_r = sqrt((r - Rg) / (Rt - Rg));
  // See Collienne to understand the mapping
  float u_mu = atan((mu + 0.15) / 1.15 * tan(1.5)) / 1.5;
  
  return texture(tex, vec2(u_mu, u_r)).rgb;
}

// Given a position r and direction mu, calculates de transmittance along the ray with
// length d. This function uses the propriety of Transmittance:
// T(a,b) = TableT(a,v)/TableT(b, v)
// r := height of starting point vect(x)
// mu := cosine of the zeith angle of vec(v). Or mu = (vec(x) * vec(v))/r
vec3 transmittance(sampler2D tex, float r, float mu, float d, float Rg, float Rt) {
  // Here we use the transmittance property: T(x,v) = T(x,d)*T(d,v) to, given a distance
  // d, calculates that transmittance along that distance starting in x (height r):
  // T(x,d) = T(x,v)/T(d,v).
  // 
  // From cosine law: c^2 = a^2 + b^2 - 2*a*b*cos(ab)
  float ri = sqrt(d * d + r * r + 2.0 * r * d * mu);
  // mu_i = (vec(d) dot vec(v)) / r_i
  //      = ((vec(x) + vec(d-x)) dot vec(v))/ r_i
  //      = (r*mu + d) / r_i
  float mui = (d + r * mu) / ri;

  // It's important to remember that we calculate the Transmittance table only for zenith
  // angles between 0 and pi/2+episilon. Then, if mu < 0.0, we just need to invert the
  // view direction and the start and end points between them, i.e., if
  // x --> x0, then x0-->x.
  // Also, let's use the property: T(a,c) = T(a,b)*T(b,c)
  // Because T(a,c) and T(b,c) are already in the table T, T(a,b) = T(a,c)/T(b,c).
  vec3 res;
  if (mu > 0.0) {
    res = transmittance(tex, r, mu, Rg, Rt) / transmittance(tex, ri, mui, Rg, Rt);
  }
  else {
    res = transmittance(tex, ri, -mui, Rg, Rt) / transmittance(tex, r, -mu, Rg, Rt);
  }
  return min(res, 1.0);
}

// Calculates Rayleigh phase function given the scattering cosine angle mu
// mu := cosine of the zeith angle of vec(v). Or mu = (vec(x) * vec(v))/r
float rayleighPhaseFunction(bool advancedModeEnabled, bool useOnlyAdvancedMie,
                            bool usePenndorfPhaseFunction, float mu)
{
    if ((advancedModeEnabled && !useOnlyAdvancedMie) || usePenndorfPhaseFunction) {
    // Penndorf
    return 0.7629 * (1.0 + 0.932 * mu * mu) * 0.25 / M_PI;
  }
  else {
    // return (3.0f / (16.0f * M_PI)) * (1.0f + mu * mu);
    return 0.0596831036 * (1.0 + mu * mu);
  }
}


// -- Calculates DHG Mie phase function given thescattering cosine angle mu --
// mu := cosine of the zeith angle of vec(v). Or mu = (vec(x) * vec(v))/r
// g1 := forward scattering 
// g2 := backward scattering
// alpha: = ratio between g1 and g2
vec3 miePhaseFunctionDHG(float mu, vec3 g1, vec3 g2, vec3 alpha) {
  // assert(advancedModeEnabled)
  // assert(!useCornettePhaseFunction)

  // Double Henyey-Greenstein phase function
  vec3 g1Sq = g1 * g1;
  vec3 g2Sq = g2 * g2;
  const float exponent = 3.0 / 2.0;

  vec3 denom1 = vec3(1.0 + g1Sq - 2.0 * g1 * mu);
  vec3 denom2 = vec3(1.0 + g2Sq - 2.0 * g2 * mu);
  vec3 d1Powered = vec3(pow(denom1.r, exponent), pow(denom1.g, exponent), pow(denom1.b, exponent));
  vec3 d2Powered = vec3(pow(denom2.r, exponent), pow(denom2.g, exponent), pow(denom2.b, exponent));

  vec3 f1 = ((vec3(1.0) - g1Sq) / d1Powered);
  vec3 f2 = ((vec3(1.0) - g2Sq) / d2Powered);
  return (alpha * f1 + (vec3(1.0) - alpha) * f2) * 0.25 / M_PI;
}

// Calculates Mie phase function given the scattering cosine angle mu
// mu   := cosine of the zeith angle of vec(v). Or mu = (vec(x) * vec(v)) / r
// mieG := mie phase function value
vec3 miePhaseFunctionBasic(float mu, float mieG) {
  float mieG2 = mieG * mieG;
  float cornette = 0.1193662072 * (1.0 - mieG2) *
    pow(1.0 + mieG2 - 2.0 * mieG * mu, -1.5) * (1.0 + mu * mu) / (2.0 + mieG2);
  return vec3(cornette);
}

vec3 miePhaseFunction(bool advancedModeEnabled, bool useCornettePhaseFunction,
                      float mu, float mieG, vec3 g1, vec3 g2, vec3 alpha)
{
  if (advancedModeEnabled && !useCornettePhaseFunction) {
    return miePhaseFunctionDHG(mu, g1, g2, alpha);
  }
  else {
    return miePhaseFunctionBasic(mu, mieG);
  }
}

// Calculates Rayleigh Scattering Coefficients given the real part of the refractive index
// and the wavelength of the incident light.
// lambda := wavelength of the incident light (680, 550, 440) nm in our case
// m := Real(n(lambada))
vec3 scatteringCoefficientRayleigh(bool advancedModeEnabled, bool useOnlyAdvancedMie,
                                   vec3 lambda, vec3 n_real_rayleigh,
                                   float deltaPolarizability, float N_rayleigh,
                                   vec3 betaRayleigh)
{
  if (advancedModeEnabled && !useOnlyAdvancedMie) {
    vec3 lambda4 = lambda * lambda * lambda * lambda;
    vec3 m2 = n_real_rayleigh * n_real_rayleigh;
    vec3 alpha = (m2 - vec3(1.0));
    float fDelta = (6.0 + 3.0 * deltaPolarizability) / (6.0 - 7.0 * deltaPolarizability);
    const float mTokm = 1000.0;
    return mTokm * ((8.0 * M_PI * M_PI * M_PI * alpha * alpha) / (3.f * N_rayleigh * lambda4)) * fDelta;
  }
  else {
    return betaRayleigh;
  }
}

// Calculates Rayleigh absorption Coefficients for a given type of molecule, given the
// refractive index and the wavelength of the incident light.
// lambda := wavelength of the incident light (680, 550, 440) nm in our case
vec3 absorptionCoefficientRayleigh(vec3 lambda, vec3 n_real_rayleigh,
                                   vec3 n_complex_rayleigh,
                                   float radius_abs_molecule_rayleigh,
                                   float N_rayleigh_abs_molecule)
{
  // n = m + ki => n^2 = (m^2 - k^2) + (2mk)i
  // n^2 - 1 = (m^2 - k^2 - 1) + (2mk)i = a
  // n^2 + 2 = (m^2 - k^2 + 2) + (2mk)i = b
  // a/b = [a x b*]/[b x b*], where b* is the conjugate of b
  vec3 n2_real    = (n_real_rayleigh * n_real_rayleigh) - (n_complex_rayleigh * n_complex_rayleigh);
  vec3 n2_complex = (2.0 * n_real_rayleigh * n_complex_rayleigh);
  
  vec3 a_real    = n2_real - vec3(1.0);
  vec3 b_real    = n2_real + vec3(2.0);
  vec3 a_complex = n2_complex;
  vec3 b_complex = n2_complex;
  vec3 b_complex_conjugate = -b_complex;
  
  //vec3 ab_conjugate_real    = (a_real * b_real) - (a_complex * b_complex_conjugate);
  vec3 ab_conjugate_complex = (a_real * b_complex_conjugate + a_complex * b_real);
  vec3 bb_conjugate = (b_real * b_real) + (b_complex * b_complex);
  vec3 imgPart = ab_conjugate_complex / bb_conjugate;

  float r3 = radius_abs_molecule_rayleigh * radius_abs_molecule_rayleigh * radius_abs_molecule_rayleigh;
  return ((8.0 * M_PI * M_PI * N_rayleigh_abs_molecule * r3) / lambda) * imgPart;
}

vec3 extinctionCoefficientRayleigh(bool advancedModeEnabled, bool useOnlyAdvancedMie,
                                   vec3 lambda, vec3 n_real_rayleigh,
                                   float deltaPolarizability, float N_rayleigh,
                                   vec3 betaRayleigh, vec3 n_complex_rayleigh,
                                   float radius_abs_molecule_rayleigh,
                                   float N_rayleigh_abs_molecule)
{
  vec3 scatter = scatteringCoefficientRayleigh(advancedModeEnabled, useOnlyAdvancedMie,
    lambda, n_real_rayleigh, deltaPolarizability, N_rayleigh, betaRayleigh);

  if (advancedModeEnabled && !useOnlyAdvancedMie) {
    vec3 absorption = absorptionCoefficientRayleigh(lambda, n_real_rayleigh,
      n_complex_rayleigh, radius_abs_molecule_rayleigh, N_rayleigh_abs_molecule);
    return scatter + absorption;
  }
  else {
    return scatter;
  }  
}

// Calculates Mie Scattering Coefficients given the wavelength of the incident light
// lambda := wavelength of the incident light (680, 550, 440) nm in our case
vec3 scatteringCoefficientMie(bool advancedModeEnabled, vec3 lambda, float jungeExponent,
                              float turbidity, vec3 Kappa, vec3 betaMieScattering)
{
  if (advancedModeEnabled) {
    vec3 tmp = (2.0 * M_PI / lambda);
    float exponent = jungeExponent - 2.0;
    vec3 lambdaJunge = vec3(
      pow(tmp.x, exponent), 
      pow(tmp.y, exponent), 
      pow(tmp.z, exponent)
    );
    return 0.434 * (0.65 * turbidity - 0.65) * 1E-16 * M_PI * Kappa * lambdaJunge;
  }
  else {
    return betaMieScattering;
  }
}

vec3 mieExtinctionEfficiency(float mean_radius_particle_mie,
                             vec3 n_real_mie, vec3 lambda, vec3 n_complex_mie,
                             vec3 betaMieExtinction)
{
  vec3 rho = 4.0 * M_PI * mean_radius_particle_mie * (n_real_mie - vec3(1.0)) / lambda;
  vec3 tanBeta = n_complex_mie / (n_real_mie - vec3(1.0));
  vec3 beta = atan(tanBeta);
  vec3 expRhoTanBeta = exp(-rho * tanBeta);
  vec3 cosBetaOverRho = cos(beta) / rho;
  vec3 cosBetaOverRho2 = cosBetaOverRho * cosBetaOverRho;
  return 2.0 - 4.0 * expRhoTanBeta * cosBetaOverRho * sin(rho - beta)
        - 4.0 * expRhoTanBeta * cosBetaOverRho2 * cos(rho - 2.0 * beta)
        + 4.0 * cosBetaOverRho2 * cos(2.0 * beta);
}


// Calculates the Mie Absorption Coefficients for dust (of particles with same radius size
// and type)
vec3 extinctionCoefficientMie(bool advancedModeEnabled, float mean_radius_particle_mie,
                             vec3 n_real_mie, vec3 lambda, vec3 n_complex_mie,
                             vec3 betaMieExtinction, float N_mie)
{
  if (advancedModeEnabled) {
    vec3 mieExtEff = mieExtinctionEfficiency(mean_radius_particle_mie,
      n_real_mie, lambda, n_complex_mie, betaMieExtinction);
    return mieExtEff * N_mie * M_PI * (mean_radius_particle_mie * mean_radius_particle_mie);
  }
  else {
    return betaMieExtinction;
  }
}

// Given the height rm view-zenith angle (cosine) mu, sun-zenith angle (cosine) muSun and
// the angle (cosine) between the vec(s) and vec(v), nu, we access the 3D textures and
// interpolate between them (r) to find the value for the 4D texture.
// r := height of starting point vect(x)
// mu := cosine of the zeith angle of vec(v). Or mu = (vec(x) * vec(v))/r
// muSun := cosine of the zeith angle of vec(s). Or muSun = (vec(s) * vec(v))
// nu := cosine of the angle between vec(s) and vec(v)
vec4 texture4D(sampler3D table, float r, float mu, float muSun, float nu, float Rg,
               int samplesMu, float Rt, int samplesR, int samplesMuS,
               int samplesNu)
{

  /*
  float r2 = r * r;
  float Rg2 = Rg * Rg;
  float Rt2 = Rt * Rt;
  float rho = sqrt(r2 - Rg2);
  float rmu = r * mu;
  float delta = rmu * rmu - r2 + Rg2;

  vec4 cst = rmu < 0.0 && delta > 0.0 ?
    vec4(1.0, 0.0, 0.0, 0.5 - 0.5 / float(samplesMu)) :
    vec4(-1.0, Rt2 - Rg2, sqrt(Rt2 - Rg2), 0.5 + 0.5 / float(samplesMu));

  float u_r = 0.5 / float(samplesR) + rho / sqrt(Rt2 - Rg2) * (1.0 - 1.0 / float(samplesR));
  float u_mu = cst.w + (rmu * cst.x + sqrt(delta + cst.y)) / (rho + cst.z) * (0.5 - 1.0 / samplesMu);
  float u_mu_s = 0.5 / float(samplesMuS) +
    (atan(max(muSun, -0.1975) * tan(1.386)) * 0.9090909090909090 + 0.74) * 0.5 * (1.0 - 1.0 / float(samplesMuS));
  float t = (nu + 1.0) / 2.0 * (float(samplesNu) - 1.0);
  float u_nu = floor(t);
  t = t - u_nu;

  vec4 v1 = texture(table, vec3((u_nu + u_mu_s) / float(samplesNu), u_mu, u_r));
  vec4 v2 = texture(table, vec3((u_nu + u_mu_s + 1.0) / float(samplesNu), u_mu, u_r));
  return mix(v1, v2, t);
  */

  float r2     = r * r;
  //float H      = sqrt(Rt2 - Rg2);
  float Rg2 = Rg * Rg;
  float rho    = sqrt(r2 - Rg2);
  float rmu    = r * mu;
  float delta  = rmu * rmu - r2 + Rg2;
  float invSamplesMu = 1.0 / float(samplesMu);
  float invSamplesR = 1.0 / float(samplesR);
  float invSamplesMuS = 1.0 / float(samplesMuS);
  float invSamplesNu = 1.0 / float(samplesNu);
  // vec4 cst     = rmu < 0.0f && delta > 0.0f ?
  //   vec4(1.0f, 0.0f, 0.0f, 0.5f - 0.5f / float(SAMPLES_MU)) :
  //   vec4(-1.0f, H * H, H, 0.5f + 0.5f / float(SAMPLES_MU));

  float Rt2 = Rt * Rt;
  float H = sqrt(Rt2 - Rg2);
  float H2 = Rt2 - Rg2;

  vec4 cst     = rmu < 0.0 && delta > 0.0 ?
    vec4(1.0, 0.0, 0.0, 0.5 - 0.5 * invSamplesMu) :
    vec4(-1.0, H2, H, 0.5 + 0.5 * invSamplesMu);

  //float u_r    = 0.5f / float(SAMPLES_R) + rho / H * (1.0f - 1.0f / float(SAMPLES_R));
  float u_r    = 0.5f * invSamplesR + rho / H * (1.0 - invSamplesR);
  //float u_mu   = cst.w + (rmu * cst.x + sqrt(delta + cst.y)) / (rho + cst.z) * (0.5f - 1.0f / float(SAMPLES_MU));
  float u_mu   = cst.w + (rmu * cst.x + sqrt(delta + cst.y)) / (rho + cst.z) * (0.5 - invSamplesMu);
  // float u_mu_s = 0.5f / float(SAMPLES_MU_S) +
  //   (atan(max(muSun, -0.1975) * tan(1.26f * 1.1f)) / 1.1f + (1.0f - 0.26f)) * 0.5f * (1.0f - 1.0f / float(SAMPLES_MU_S));
  float u_mu_s = 0.5 * invSamplesMuS +
    (atan(max(muSun, -0.1975) * tan(1.386)) * 0.9090909090909090 + (0.74)) * 0.5 * (1.0 - invSamplesMuS);
  float lerp = (nu + 1.0) / 2.0 * (float(samplesNu) - 1.0);
  float u_nu = floor(lerp);
  lerp = lerp - u_nu;

  // return texture(table, vec3((u_nu + u_mu_s) / float(SAMPLES_NU), u_mu, u_r)) * (1.0f - lerp) +
  //   texture(table, vec3((u_nu + u_mu_s + 1.0f) / float(SAMPLES_NU), u_mu, u_r)) * lerp;
  
  return texture(table, vec3((u_nu + u_mu_s) * invSamplesNu, u_mu, u_r)) * (1.0 - lerp) +
    texture(table, vec3((u_nu + u_mu_s + 1.0) * invSamplesNu, u_mu, u_r)) * lerp;
}
